%% tclp__declarations.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__declarations.pl,v 1.12 2004/11/08 15:42:27 coquery Exp $

%% Memorisation of user type declarations 

:- module(tclp__declarations, [untyped_term/1, get_declaration/2,
                               get_declarationG/2, get_declaration_enum/2,
                               check_declarations/0]).

:- use_module(library(terms)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__type_order)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__handle_goals)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__files)).
:- use_module(tclp(tclp__utils)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Term declaration
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% type declarations are stored in some dynamic predicates
%% type information  are of the form
%% types_data(ListOfListOfTypes, Goal) where the first argument is a list
%% of all possible types organized has [ ResultType | ArgsType ].

:- dynamic declared_functor_types/3, declared_functor_typesG/3.

%% change the type of a functor
set_declared_functor_types(F,N,TypeData) :-
        retractall(declared_functor_types(F,N,_)),
        asserta(declared_functor_types(F,N,TypeData)).

set_declared_functor_typesG(F,N,TypeData) :-
        retractall(declared_functor_typesG(F,N,_)),
        asserta(declared_functor_typesG(F,N,TypeData)).

%% adds a declaration to all the functors types when its not already in
%% the fact base.
add_declaration(F/N, NewArgsType, NewResultType, NewGoal) :-
        verbose(debug_declarations, ['Entering add_declaration ',F/N]),
        (
          declared_functor_types(F,N,types_data(CurrentTypes, CurrentGoal))
        ;
          CurrentTypes = [], CurrentGoal = true
        ),
        verbose(debug_declarations, ['found olds ', CurrentTypes,
                                     ' ', CurrentGoal]),
        (
          ( find_variant([NewResultType|NewArgsType],CurrentTypes);
              length(CurrentTypes,N),N>0,option(disable_overloading) )
        ->
          %Types = CurrentTypes, Goal = CurrentGoal
          % Dont change anything
          verbose(debug_declarations, 'ok: no change'),
          true
        ;
          Types =[ [ NewResultType|NewArgsType] | CurrentTypes],
          Goal = (CurrentGoal, NewGoal),
          verbose(debug_declarations, [ 'Changing ',F/N,
                                        ' from ', CurrentTypes,
                                        ' to ', Types ]),
          set_declared_functor_types(F,N,types_data(Types,Goal)),
          verbose(debug_declarations, 'change ok')
        ).

%% find_variant(Term,List) looks in list for a variant of Term
find_variant(Term,[X|R]) :-
        variant(Term,X) -> true ; find_variant(Term,R).

%% Replace the variables by atoms representing their names
generalize(X,Name,TermData) :-
        var(X),!,
        verbose(debug_declarations, ['generalize variable ',X]),
        (
          variable_name(X,TermData,Name),
          add_type_variable_constant(Name)
        ;
          print_error(TermData, ['name failed for ',X])
        ).
generalize(Term, GTerm, TermData) :-
        verbose(debug_declarations, ['generalize term ',Term]),
        Term =.. [ Head | Args ],
        generalize_list(Args, GArgs, TermData),
        GTerm =.. [ Head | GArgs ].
generalize_list([],[],_) :- !.
generalize_list([X|Rest],[Gen|GRest], TermData) :- !,
        generalize(X,Gen,TermData),
        generalize_list(Rest, GRest, TermData).

%% adds a declaration to all the functors types (generalized)
add_declarationG(F/N, NewArgsType, NewResultType, NewGoal, TermData) :-
        verbose(debug_declarations, ['Entering add_declarationG ',F/N]),
        generalize([NewResultType|NewArgsType],GTypes,TermData),
        verbose(debug_declarations, 'generalize ok'),
        (
          declared_functor_typesG(F,N,types_data(CurrentTypes, CurrentGoal))
        ;
          CurrentTypes = [], CurrentGoal = true
        ),
        verbose(debug_declarations, ['found olds ', CurrentTypes,
                                     ' ', CurrentGoal]),
        (
          find_variant(GTypes,CurrentTypes)
        ->
          % Types = CurrentTypes, Goal = CurrentGoal
          % no change needed
          verbose(debug_declarations, 'ok : no change'),
          true
        ;
          Types =[ GTypes | CurrentTypes],
          Goal = (CurrentGoal, NewGoal),
          verbose(debug_declarations, [ 'Changing ',F/N,
                                        ' from ', CurrentTypes,
                                        ' to ', Types ]),
          set_declared_functor_typesG(F,N,types_data(Types,Goal)),
          verbose(debug_declarations, 'change ok')
        ).

%% get all the types for a functor
get_declaration(F/N, Types) :-
        declared_functor_types(F,N,types_data(Types, Goal)),!,
        call(Goal).

%% get all the types for a functor
get_declarationG(F/N, Types) :-
        declared_functor_typesG(F,N,types_data(Types, Goal)),!,
        call(Goal).

%% enumerates all the types for a functor
get_declaration_enum(F/N, Type) :-
        declared_functor_types(F,N,types_data(Types,_)),
        member(Type,Types).

%% checks if all declarations contain valid types
check_declarations :-
        declared_functor_types(F,N,types_data(Types,_)),
        %write(F/N/Types),nl,
        %(check_types(F,N,Types)->write(ok),nl;write(fail),nl),
        check_types(F,N,Types),
        (option(print_types) ->
            (length(Types,Len), Len=1 -> ARE=' is '; ARE=' are '),
            verbose(always,['type of ',F/N,ARE,Types]) ;
            true ),
        fail
        ;
        true.

%% checks the types, F/N is the tested functor
check_types(_,_,[]).
check_types(F,N,[T|L]) :- !,
        %write('check_types '),write([T|L]),nl,
        check_type(F,N,T),
        check_types(F,N,L).
%check_types(F,N,X) :- write(error/F/N/X),nl,fail.

check_type(_,_,[]).
check_type(F,N,[T|L]) :- %!,
        (type_exists(T) ->
        check_type(F,N,L) ;
        print_error(['The type ',T,' used in type declaration for ',F/N,' does not exist']),fail).
%check_type(F,N,X) :- write(error2/F/N/X),nl,fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Untyped terms handling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Some predicates or terms may untyped , for example if trying to typecheck
%% them would conduct to a type error (e.g. mode/1)
%% We allow to add some 'untyped' terms. they may have any type and their
%% argument are not type checked (see tclp__type_checker)
%% We store them using the dynamic predicate untyped_term/1

:- dynamic untyped_term/1.

%% 1)
%% :- untyped [ f1/n1 , f2/n2 , ... ].

%% We define an operator
:- op(999, fx, [ untyped ]).

%% This predicate adds terms to the untyped data base
untyped [].
untyped [ A | Rest ] :-
        untyped A,
        untyped Rest.
untyped F/N :-
        asserta(untyped_term(F/N)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Goals used for declarations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% functor's types declarations

%% 1)
%% :- typeof Term is Type.

%% We define the typeof operator
:- op(999, fx, [ typeof ]).
%% and the with for constraints
:- op(998, xfx, [ with ]).
                  
%% typeof ATermWithTypes is AType is the declaration used to define types
%% of terms.
typeof((ATermWithTypes is AType) with Constraints, TermData) :-
        args_location(TermData,[TTermData,CTermData]),
        check_typeof_constraints(Constraints,CTermData),!,
        add_type_declaration(ATermWithTypes,AType, TTermData, Constraints).
typeof(ATermWithTypes is AType, TermData) :-
        add_type_declaration(ATermWithTypes,AType, TermData, true).
typeof(_,TermData) :-
        print_error(TermData,[ 'Error in typeof declaration' ]).

add_type_declaration(ATermWithTypes,AType, TermData, OtherGoal) :-
        verbose(debug_declarations, [ 'Entering ',
                                      (typeof ATermWithTypes is AType)]),
        functor(ATermWithTypes, F, N),
        ATermWithTypes =.. [ F | ArgsType ],
        %% handle atomic terms (which also have type atom)
        (
          atom(ATermWithTypes),
          builtin_type(atom,Atom),
          AType \= Atom,
          \+ option(disable_overloading)
        ->
          typeof(ATermWithTypes is Atom, TermData)
        ;
          true
        ),
        %% Add free variable to the store whenever we use an instance of
        %% the type scheme
        Goal = (fresh_term(ArgsType), fresh_term(AType), OtherGoal),
        %% add the declaration to the data base
        add_declaration(F/N, ArgsType, AType, Goal),
        %(
        %  nonvar(AType)%, AType = pred
        %->
        add_declarationG(F/N, ArgsType, AType, Goal, TermData),
        %;
        %  true
        %),
        verbose(debug_declarations,'typeof suceeded').

%%%%%%%%%%%%%%%%%%%%%%
%% Check that the constraints are only subtype constraints in typeof
check_typeof_constraints((X,Y),TermData) :- !,
        args_location(TermData,[XTermData,YTermData]),
        check_typeof_constraints(X,XTermData),
        check_typeof_constraints(Y,YTermData).
check_typeof_constraints(':<'(_,_),_) :- !.
check_typeof_constraints('='(_,_),_) :- !.
check_typeof_constraints(C,TermData) :-
        var(C)
        ->
        variable_name(C,TermData,Name),
        print_error(TermData,['Variable ',Name,' used as a constraint'])
        ;
        functor(C,F,N),
        print_error(TermData,[F/N,' is not allowed in typeof declarations']).

%%%%%%%%%%%%%%%%%%%%%%
%% Type declarations


%% 1)
%% :- type Type is [ Term, Term, ... ]

%% add_all_functors(List, Type).
%% adds all the functors in List with Type as their result type
add_all_functors(_,[],_).
add_all_functors(Type, [Term|Rest],TermData) :-
        typeof(Term is Type,TermData),
        add_all_functors(Type, Rest,TermData).

%% Operator definition
:- op(999, fx, [ type ]).

%% type AType is [ ATermWithTypes, ATermWithTypes, ... ]
%% declares AType as a type and the type of functors ATermWithTypes is AType
type(Type is ListOfTerms,TermData) :- !,
        add_type(Type),
        verbose(debug_declarations, 'add_type ok'),
        add_all_functors(Type, ListOfTerms,TermData).
        
%% 2)
%% :- type Type.
%% :- type Cons/NArgs.

%% type Type introduce a type into the data base.
%% An alternative declaration is Cons/NArgs
type(Cons/NArgs,_) :-
        nonvar(Cons), nonvar(NArgs), !,
        add_type(Cons, NArgs).
type(Type,_) :-
        add_type(Type).

%% 3)
%% :- order Type1 < Type2

%% Operator definition
:- op(999, fx, [ order ]).
:- op(999, fx, [ subtype ]).

%% order T1 < T2 adds the constraint that the constructor T1 must be
%% smaller or equal to constructor T2 and also adds the correspondance
%% constraint on their arguments.
order T1 < T2 :-
       add_order(T1, T2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% File inclusion
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tclp__include(File,Location) :-
        catch(load_file(File),E,
              (get_location(Location,L),write('In '), write(L),nl,throw(E))).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Clauses regognition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% A shortcut for defining binary clause symbols (such as +: for fd)
tclp__add_clause_op(Op,Type) :-
        Goal = (
                 C =.. [ Op , H , B ],
                 user:args_location(L,[HL,BL])
               ),
        tclp__handle_goals:tclp__add_clause_layout(C,L,[H-HL-Type],
                                                   [B-BL],Goal).
tclp__add_clause_op(Op) :-
        builtin_type(predicate,Pred),
        tclp__add_clause_op(Op,Pred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Compatibility
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(1150, fx, [ val, pred , force_type ]).
enable_backward_compatibility :-
        hook_val_back, hook_pred_back, hook_order_back, hook_type_back.
tclp__convert_bool(X,X) :- var(X),!.
tclp__convert_bool(bool,Pred) :- !, builtin_type(predicate, Pred).
tclp__convert_bool([],[]) :- !.
tclp__convert_bool([X|L],[Y|R]) :-
        tclp__convert_bool(X,Y),
        tclp__convert_bool(L,R).
tclp__convert_bool(Term,NewTerm) :-
        Term =.. [ H | Args ],
        tclp__convert_bool(Args,NewArgs),
        NewTerm =.. [ H | NewArgs].

add_val_back(TypedTerm,Type,Location) :-
        tclp__convert_bool(TypedTerm,NTypedTerm),
        tclp__convert_bool(Type, NType),
        typeof(NTypedTerm is NType,Location).
add_pred_back(PredType, Location) :-
        tclp__convert_bool(PredType, NPredType),
        builtin_type(predicate,Pred),
        typeof(NPredType is Pred, Location).
add_order_back(A) :-
        tclp__convert_bool(A,NA),
        order NA.
add_type_back(A,L) :-
        tclp__convert_bool(A,NA),
        type(NA,L).

hook_val_back :-
        tclp__executable((val TypedTerm : Type)),
        tclp__add_hook((val TypedTerm : Type), Location,
                       tclp__declarations:add_val_back(TypedTerm, Type,
                                                       Location)
                      ).

hook_pred_back :-
        tclp__executable((pred PredType)),
        tclp__add_hook((pred PredType), Location,
                       tclp__declarations:add_pred_back(PredType, Location)
                      ).

hook_order_back :-
        tclp__add_hook((order A), tclp__declarations:add_order_back(A)).

hook_type_back :-
        tclp__add_hook((type A), L, tclp__declarations:add_type_back(A,L)).

reinit :-
        retractall(declared_functor_types(_,_,_)),
        retractall(declared_functor_typesG(_,_,_)),
        retractall(untyped_term(_)),
        tclp__executable(untyped _),
        tclp__add_hook(untyped A, tclp__declarations:untyped(A)),
        tclp__executable(typeof _),
        tclp__add_hook(typeof(B),TData,
                       tclp__declarations:typeof(B,TData)),
        tclp__executable(type _),
        tclp__add_hook(type D,TData2, tclp__declarations:type(D,TData2)),
        tclp__executable(order _),
        tclp__add_hook(order E, tclp__declarations:order(E)),
        tclp__executable(subtype _),
        tclp__add_hook(subtype E, tclp__declarations:order(E)),
        tclp__executable(tclp__include(_)),
        tclp__add_hook(tclp__include(F), Loc,
                       tclp__declarations:tclp__include(F,Loc)),
        tclp__executable(tclp__define_clause(_,_,_,_)),
        tclp__executable(tclp__define_clause(_,_,_,_,_)),
        tclp__add_hook(tclp__define_clause(C1,L1,H1,B1),
                       tclp__handle_goals:tclp__add_clause_layout(C1,L1,H1,B1)),
        tclp__add_hook(tclp__define_clause(C2,L2,H2,B2,Co2),
                       tclp__handle_goals:tclp__add_clause_layout(C2,L2,H2,B2,Co2)),
        tclp__executable(tclp__define_clause_op(_)),
        tclp__add_hook(tclp__define_clause_op(Op),
                       tclp__declarations:tclp__add_clause_op(Op)),
        tclp__executable(tclp__define_clause_op(_,_)),
        tclp__add_hook(tclp__define_clause_op(Op2,Type),
                       tclp__declarations:tclp__add_clause_op(Op2,Type)).

:- reinit.
