%% tclp__type_checker.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__type_checker.pl,v 1.17 2004/04/22 11:34:10 coquery Exp $

%% The type checker used by tclp

:- module(tclp__type_checker,[type_check_clauses/1, var_type_reinit/0]).

:- use_module(tclp(tclp__utils)).
:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__handle_goals)).
:- use_module(tclp(tclp__declarations)).
:- use_module(tclp(tclp__functor_types)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__reader)).
:- use_module(tclp(tclp__typing_environment)).
:- use_module(library(chr)).
:- use_module(library(lists)).

%% Verbose mode : debug_type_checker

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A small chr constraint handler for the type of variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
handler variable_checker.
%% the constraint X::T says that the type of X is represented by type
%% constraints over the type variable T.
%% :- pred '::'(term,type_term).
operator(100,xfx,'::').
constraints '::'/2, var_type_reinit/0. 

collect1    @ var_type_reinit \ _::_ <=> true.
collectself @ var_type_reinit <=> true.

%% rules
%% when a term has two types, we unify them
type_ident @ V::T1 \ V::T2 <=> T1=T2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% type checking predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% '' is Type is bottom, Type-'' otherwise
erase_bottom(X,''):- (builtin_type(bot,Y)->Y==X),!.
erase_bottom(X,X-'').

%% This predicate type-checks a given term with its location infos
%% :- pred type_check(term,term_data, type_term).
type_check(X, TermData, Type) :-
        var(X), !, fresh(Type), 
        (
          verbose(debug_type_checker,TermData,['Testing ',X,' :: ',Type]),
          X::Type,
          verbose(debug_type_checker,[X,' :: ',Type,' ok']),
          verbose(debug_type_checker,['adding variable ',X,' ',
                                      TermData,' ',Type,' ...']),
          tclp__typing_environment:add_variable(X,TermData,Type),
          verbose(debug_type_checker,['added variable ',X,' ',
                                      TermData,' ',Type])
        ->
          true
        ;
          (
            verbose(debug_type_checker,'starting error message build'),
            chr:findall_constraints(X,X::_,[(X::XType)#_]),
            verbose(debug_type_checker,'find constraints ok'),
            type_info(XType, XLB-XUB),
            erase_bottom(XLB,XLB2),
            verbose(debug_type_checker,'XTypeInfo ok'),
            type_info(Type, TLB-TUB),
            erase_bottom(TLB,TLB2),
            verbose(debug_type_checker,'TypeInfo ok'),
            verbose(debug_type_checker,[ X,' ', TermData]),
            variable_name(X,TermData,Name),
            verbose(debug_type_checker,'Variable name ok'),
            print_env(TermData),
            verbose(debug_type_checker,'Print env ok'),
            print_error(TermData,
                        [' Incompatible types for ',Name,' : ',
                         XLB2,XUB,' and ',TLB2,TUB])
          ;
            print_error('error in type_check failed')
          )
        ).

type_check(Term, TermData, Type) :-
        verbose(debug_type_checker,['checking ',Term]),
        functor(Term, Head, Arity),
        (
          untyped_term(Head/Arity)
        ->
          fresh(Type), verbose(debug_type_checker,'untyped')
        ;
          Term =.. [ Head | Args ],
          length(Args,N),
          length(ArgsTypes,N),!,
          args_data(TermData, ArgsData,N),
          verbose(debug_type_checker,'args data ok'),
          apply(Head/Arity, ArgsTypes, Type, TermData),
          verbose(debug_type_checker,'apply ok'),
          type_check_args(Args, ArgsData, ArgsTypes),
          verbose(debug_type_checker,'args check ok'),
          verbose(debug_type_checker,['checking ',Term,' ok'])
        ),!.

%% This version of the type checker takes care of type inference for predicates
%% :- pred type_check(term, term_data, type_term,
%%                    list(pair(functor,list(type_term)))).
type_check(X, TermData, Type, _) :-
        var(X), !,
        type_check(X, TermData, Type).
type_check(Term, _, Type, _) :-
        functor(Term, Head, Arity),untyped_term(Head/Arity),!,
        fresh(Type).
type_check(Term, TermData, Type, InferedPreds) :-
        functor(Term, Head, Arity),
        (
          once((Term =.. [ Head | Args ],
                length(Args,N),
                length(ArgsTypes,N),
                args_data(TermData, ArgsData,N)
               )),
          (
            member((Head/Arity)-FunctorArgsTypes-FunctorType, InferedPreds)
          ->
            once((%builtin_type(predicate, FunctorType),
                  %fresh_term(ArgsTypes),
                  link_args(Head/Arity,ArgsTypes, FunctorArgsTypes, ArgsData),
                  %fresh(Type),
		  FunctorType = Type
                 ;
                  print_env(TermData),
                  print_error(TermData,
                              [ 'Could not associate a type to ',
                                Head/Arity ])
                 ))
          ;
            verbose(debug_type_checker,['apply ',Head/Arity]),
            apply(Head/Arity, ArgsTypes, Type, TermData),
            verbose(debug_type_checker,['apply ',Head/Arity,' ok'])
          ),
          type_check_args(Args, ArgsData, ArgsTypes, InferedPreds),
	  verbose(debug_type_checker,['type check ',Term,' ok'])
        ),!.
        

%% This version can be used to type check heads with polymorphic types.
%% Should be used only on known heads
%% Works only on decomposable terms
type_checkG(Term, TermData, Type, InferedPreds) :-
        functor(Term, Head, Arity),!,
        Term =.. [ Head | Args ],
        length(Args,N),
        length(ArgsTypes,N),
        args_data(TermData, ArgsData,N),!,
        verbose(debug_type_checker,['applyG ...']),
        applyG(Head/Arity, ArgsTypes, Type, TermData),
        verbose(debug_type_checker,['applyG OK']),
        type_check_args(Args, ArgsData, ArgsTypes, InferedPreds).
        

%% This version can be used to type check heads with polymorphic types.
%% Should be used only on known heads
%% Works only on decomposable terms
type_checkG(Term, TermData, Type) :-
        functor(Term, Head, Arity),!,
        Term =.. [ Head | Args ],
        length(Args,N),
        length(ArgsTypes,N),
        args_data(TermData, ArgsData,N),!,
        applyG(Head/Arity, ArgsTypes, Type, TermData),
        type_check_args(Args, ArgsData, ArgsTypes).
        

%% This predicate type-checks a list of terms
%% :- pred type_check_args(list(term), list(term_data), list(type_term)).
type_check_args([],[],[]).
type_check_args([Term|Tms], [TermData|Ds], [Type|Tps]) :-
        type_check(Term,TermData,Type),
        type_check_args(Tms,Ds,Tps).

%% This version of the type checker takes care of type inference for predicates
%% :- pred type_check_args(list(term), list(term_data), list(type_term),
%%                         list(pair(functor,list(type_term)))).
type_check_args([],[],[],_).
type_check_args([Term|Tms], [TermData|Ds], [Type|Tps], InferedPreds) :-
        type_check(Term,TermData,Type, InferedPreds),
        type_check_args(Tms,Ds,Tps, InferedPreds).

%% This predicates links arguments
%% :- pred link_args(atom, list(type_term), list(type_term), list(term_data)).
%% :- pred link_args(atom, list(type_term), list(type_term), list(term_data), int).
link_args(FunctorName, ArgsTypes, FunctorArgsTypes, ArgsData) :-
        link_args(FunctorName, ArgsTypes, FunctorArgsTypes, ArgsData, 1).
link_args(_,[],[],[],_).
link_args(FunctorName, [AType|ATs], [FType|FTs], [AData|ADs], N) :-
        AType :< FType, N1 is N+1, link_args(FunctorName, ATs, FTs, ADs, N1)
        ;
        line_number(AData, Number),
        type_info(AType, ATypeInfo),
        type_info(FType, FTypeInfo), 
        throw(uncompatible_arg_type(Number, FunctorName, N, ATypeInfo,
                                    FTypeInfo))
        ;
        print_error('printing error failed in link_arg').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Type check of heads
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% This predicate tests if in the list of scheme that was given, there was one
%% that can be used as a predicate
has_a_pred_typing(Types,Pred) :-
        %builtin_type(predicate, Pred),!,
        member([ RType | _ ], Types),
        RType :< Pred, !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Type check one group of clauses
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% gets the predicate spec
get_pred_spec(Module:V,Module:V) :- % bug fix when ':' is not native
                                    % to the dialect
        var(V),!.
get_pred_spec(Module:Head,Module:Spec) :- !,
        get_pred_spec(Head,Spec).
get_pred_spec(Head,F/N) :- functor(Head,F,N).

%% gets the predicates specs
get_preds_specs([],[]).
get_preds_specs([H-_-_|L],[S|R]) :-
        get_pred_spec(H,S),
        get_preds_specs(L,R).

%% gets the predicate defined by the given clause
get_clause_preds(Phrase-Loc,Specs) :-
        tclp__clause(Phrase,Loc,Heads,_),!,
        (
          Heads = [] ->
          Specs = [directive]
        ;
          get_preds_specs(Heads,Specs)
        ).

%% This predicate type checks a list of clauses
type_check_clauses([]).
type_check_clauses([ Clause | _ ]) :-
        type_check_one_clause(Clause),fail.
type_check_clauses([ _ | Rest ]) :- 
        type_check_clauses(Rest).

%%%%%%%%%
%% Type check

%% Type checking one clause
type_check_one_clause(Clause) :-
        multi_functor_reinit,
        var_type_reinit,
        types_reinit,
        init,        
        Clause = Phrase-TermData, get_clause_preds(Clause,Specs),
        get_location(TermData,Location),
        verbose(actions,['Type checking ',Specs,', ',Location]),        
        chrono(actions,'type check',
               (tclp__clause(Phrase,TermData,Heads,Bodies),
                   type_check_bodies(Bodies),
                   (type_check_heads(Heads,TermData);true) )
              )
        ,!.

%% Type checks all the bodies in the list
type_check_bodies([]).
type_check_bodies([Goal-TermData|Bodies]) :-
        type_check(Goal, TermData, Type),
        builtin_type(goal, GoalType),
        once((Type :< GoalType ;
              print_env(TermData),
              print_error(TermData,
                          [ 'The body of the directive does not have type ',
                            GoalType]))),!,
        type_check_bodies(Bodies).

%% counter for number of enumeration
% :- init_counter(enums).

%% Enumerate possible types for the heads and type check them all
type_check_heads([],TermData) :- !,
        once((chrono(actions,'try reduce', try_reduce);
              print_env(TermData),
              print_error(TermData,'try reduce failed'))),
        fail.
type_check_heads([Head-TD1-Type|HDs],TermData) :-
        type_check_head_one_type(Head,TD1,Type),
        type_check_heads(HDs,TermData).

%% Choose one possible type for the head and uses it for type check
% should be trapped by tclp__clause/3
type_check_head_one_type(Head,_,_) :-
        functor(Head,F,N),untyped_term(F/N),!.
type_check_head_one_type(Head,TermData,PredType) :-
        Head =.. [ F | Args ],
        functor(Head,F,N),
        once((get_declarationG(F/N,Types) ;
              print_error(TermData,
                          ['Predicate ',
                           F/N,' is unknown, cannot be type checked']))),
        once((has_a_pred_typing(Types,PredType) ;
              %builtin_type(predicate, Pred),
              print_error(TermData,
                          ['Functor ', F/N,' does not have type ',PredType]))),
        args_data(TermData, ArgsData,N),
        !,
        member([ RType | FArgsTypes ],Types), %enumeration by backtracking
        ( RType :< PredType ->
            once((length(ArgsTypes,N),
		  %% removing value fresh ...
                  link_args(F/N,ArgsTypes,FArgsTypes, ArgsData),
                  type_check_args(Args, ArgsData, ArgsTypes)
                 ;
                  print_error(TermData,
                              [F/N,' : ',FArgsTypes,' -> ',
                               RType,' type check failed'])
                 ))
        ).
