%% tclp__typing_environment.pl
%% library used to collect and display typing environments

%% Copyright (C) 2001,2004 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__typing_environment.pl,v 1.2 2004/02/06 16:57:09 coquery Exp $

:- module(tclp__typing_environment,[print_env/1,
                                    add_variable/3]).

:- use_module(tclp(tclp__printer)).
:- use_module(tclp(tclp__clp_types)).
:- use_module(tclp(tclp__reader)).

:- use_module(library(chr)).

handler chr_typing_environment.

constraints variable_data/2, variable_phrase/2, add_variable/3.

add @ variable_data(X,_VData)  \ add_variable(X,_TD,_Type) <=> true.

add2 @ add_variable(X,TD,Type) <=>
        get_vdata(X,TD,Type,VData),
        variable_data(X,VData),
        get_location_phrase_id(TD,Id),
        variable_phrase(X,Id).

consult @ variable_data(X,D) \ variable_data(X,D2) <=> D=D2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Printing
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

print_variable_data(X,Printer) :-
        variable_data(X,vdata(cloc(Location),Name,Type)),
        term_type_up(Type,TypeValue),
        pprint(Printer,[ Name,Location,' : ',TypeValue ]).

print_env_printer(printer('%% ! ')).

print_env(Location) :-
        get_location_phrase_id(Location,Id),
        print_env_printer(Printer) -> 
        chr:find_constraint(Id,variable_phrase(X,Id),_),
        print_variable_data(X,Printer),
        fail.
print_env(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Creating vdata structures
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_vdata(X,Location,Type,vdata(cloc(LocationValue),Name,Type)) :-
        line_number(Location,Line),
        number_codes(Line,Codes),atom_codes(LName,Codes),
        atom_concat(', line ',LName,LocationValue),
        variable_name(X,Location,Name),!.